import asyncio
import typing as ty

from ...exceptions import UniqueRequiredError

try:
    import aiosqlite
except ModuleNotFoundError as err:
    raise RuntimeError(
        "You should install `aiosqlite` backend to connect to this db. "
        "Use `pip install aiosqlite`"
    ) from err

from ..base import BaseProvider, ConnectionWrapper


class AiosqliteProvider(BaseProvider[aiosqlite.Connection]):
    def __init__(self, db_path, **connection_kwargs) -> None:
        super().__init__(db_path, **connection_kwargs)
        self.connection = None
        self._lock = asyncio.Lock()

    async def create_connection(self) -> None:
        if not self.connection:
            self.connection = await aiosqlite.connect(
                self.db_path, isolation_level=None
            )

    async def close_connection(self) -> None:
        if self.connection:
            await self.connection.close()
            self.connection = None

    def ensure_connection(self):
        return ConnectionWrapper(self, self._lock)

    async def _execute(self, query, args=()):
        async with self.ensure_connection() as connection:
            return await connection.execute(query, args)

    async def _execute_insert_query(self, query, values):
        async with self.ensure_connection() as connection:
            cursor: aiosqlite.Cursor = await connection.execute(query, values)
            rows = await cursor.fetchall()
            return [row[0] for row in rows]

    async def _fetchone(self, query, args=()) -> tuple[ty.Any] | None:
        if rows := await self._fetchall(query, args):
            return rows[0]

    async def _fetchall(self, query, args=()) -> list[tuple[ty.Any]]:
        async with self.ensure_connection() as connection:
            return list(await connection.execute_fetchall(query, args))

    @staticmethod
    def modify_db_path(db_path: str) -> str:
        return db_path.removeprefix("sqlite+aiosqlite://")

    def _translate_exception(self, exception, query, params):
        if isinstance(exception, aiosqlite.IntegrityError):
            if exception.sqlite_errorname == "SQLITE_CONSTRAINT_UNIQUE":
                text = exception.args[0]
                return UniqueRequiredError(
                    query,
                    params,
                    exception,
                    field_name=text[text.rfind(": ") + 2 :],
                )
        return super()._translate_exception(exception, query, params)



import typing as ty


class DBError(Exception):
    msg: str = ""

    def __init__(
        self,
        msg: str | None = None,
        base_exc: Exception | None = None,
        **extra: ty.Any,
    ):
        self.base_exc = base_exc
        self.extra = extra
        self.msg = (msg or self.msg).format(**self.extra)

    def __str__(self) -> str:
        exc_msg = (
            f" Base exc {type(self.base_exc).__name__}: {self.base_exc}"
            if self.base_exc
            else ""
        )
        return f"{self.msg}{exc_msg}"


class QueryError(DBError):
    def __init__(
        self,
        query: str,
        params: ty.Sequence[ty.Any],
        base_exc: Exception,
        msg: str | None = None,
        **extra: ty.Any,
    ):
        msg = (msg or self.msg) + " on query `{query}` with args {params}."
        extra.update(dict(query=query, params=params))
        super().__init__(msg, base_exc, **extra)


class UniqueRequiredError(QueryError):
    msg = "Value for field `{field_name}` must be unique."


class ConnectionIsNotAccrued(DBError):
    msg = "Connection is not accrued"



from __future__ import annotations

import typing as ty
from abc import ABC

if ty.TYPE_CHECKING:
    from .models import Field


class Join[T](ABC):
    type: str

    def __init__(
        self,
        model: ty.Type[T],
        on: tuple[Field, Field],
    ):
        self.model = model
        self.on: tuple[Field, Field] = on

    def __repr__(self):
        return (
            f'{self.type} JOIN "{self.model.__name__.lower()}" ON '
            f"{self.on[0].model_name.lower()}.{self.on[0].name.lower()}="
            f"{self.on[1].model_name.lower()}.{self.on[1].name.lower()}"
        )

    __str__ = __repr__


class InnerJoin[T](Join[T]):
    type = "INNER"


class RightJoin[T](Join[T]):
    type = "RIGHT"


class LeftJoin[T](Join[T]):
    type = "LEFT"



import asyncio
import re
import typing as ty

from ...exceptions import UniqueRequiredError

try:
    import asyncpg
except ModuleNotFoundError as err:
    raise RuntimeError(
        "You should install `asyncpg` backend to connect to this db. "
        "Use `pip install asyncpg`"
    ) from err

from ..base import BaseProvider, PoolConnectionWrapper


class AsyncpgProvider(BaseProvider[asyncpg.Connection]):
    CREATE_TABLE_QUERY_TEMPLATE = (
        'CREATE TABLE IF NOT EXISTS "{table_name}" '
        "(id SERIAL PRIMARY KEY NOT NULL, {fields})"
    )
    INSERT_INTO_QUERY_TEMPLATE = (
        'INSERT INTO "{table_name}" ({fields}) VALUES {rows} RETURNING id'
    )

    DEFAULT_FIELD_TYPE = "BYTEA"
    PLACEHOLDER: ty.Callable[[int], str] = staticmethod(lambda i: f"${i}")

    def __init__(self, db_path, **connection_kwargs) -> None:
        super().__init__(db_path, **connection_kwargs)
        self.connections_pool = None
        self._pool_init_lock = asyncio.Lock()

    async def create_connection(self) -> None:
        self.connections_pool = await asyncpg.create_pool(
            self.db_path, min_size=1, max_size=5
        )

    async def close_connection(self) -> None:
        await self.connections_pool.close()

    def ensure_connection(self):
        return PoolConnectionWrapper(self, self._pool_init_lock)

    async def _execute(self, query, args=()):
        async with self.ensure_connection() as connection:
            return await connection.execute(query, *args)

    async def _execute_insert_query(self, query, values) -> list[int]:
        async with self.ensure_connection() as connection:
            rows = await connection.fetch(query, *values)
            return [row.get("id") for row in rows]

    async def _fetchone(self, query, args=()) -> tuple[ty.Any]:
        async with self.ensure_connection() as connection:
            return await connection.fetchrow(query, *args)

    async def _fetchall(self, query, args=()) -> list[tuple[ty.Any]]:
        async with self.ensure_connection() as connection:
            return await connection.fetch(query, *args)

    @staticmethod
    def modify_db_path(db_path: str) -> str:
        return re.sub(r"\+asyncpg", "", db_path)

    def _translate_exception(self, exception, query, params):
        if isinstance(exception, asyncpg.UniqueViolationError):
            if match := re.search(
                r"Key \((.+?)\)=", getattr(exception, "detail", "")
            ):
                return UniqueRequiredError(
                    query, params, exception, field_name=match.group(1)
                )
        return super()._translate_exception(exception, query, params)



from __future__ import annotations

import types as tys
import typing as ty

from .models import Field, ModelSignature, prepare_model
from .operators import InvertedField, MathOperator
from .providers import get_provider
from .tools import get_changed_attributes

if ty.TYPE_CHECKING:
    from .joins import Join
    from .operators import Operator
    from .providers import BaseProvider


class AsyncDBCore[Models]:
    """
    Main db class.
    You can implement your needs to the database in the classroom class
    or use this class directly.
    """

    signatures: dict[str, ModelSignature] = {}
    dbs: dict[str, BaseProvider] = {}
    db_names: dict[str, str] = {}

    @classmethod
    def init(
        cls, database_path: str, db_name: str = "main", **connection_kwargs
    ) -> None:
        """
        Initialize the db.
        :param database_path: Path to db.
        :param db_name: Connection name is used for multi connection.
        :param connection_kwargs: Params for connection provider.
        """
        if db_name in cls.db_names:
            raise ValueError("DB with this name already initialized")
        cls.db_names[db_name] = database_path

        generic_types = ty.get_args(tys.get_original_bases(cls)[0])[0]
        if type(generic_types) is ty.TypeVar:
            raise NotImplementedError("")
        elif type(generic_types) in {ty.Union, tys.UnionType}:
            models = ty.get_args(generic_types)
        else:
            models = [generic_types]

        if not cls.signatures:
            cls.signatures = {
                model.__name__: prepare_model(model) for model in models
            }

        cls.dbs[db_name] = get_provider(database_path)(
            database_path, **connection_kwargs
        )

    @classmethod
    async def close_connections(cls) -> None:
        """
        Closes all connections.
        """
        for provider in cls.dbs.values():
            await provider.close_connection()

    def __init__(self, db: str = "main"):
        if not self.dbs:
            raise RuntimeError("DB is not initialized")
        if not (provider := self.dbs.get(db)):
            raise ValueError(f"DB `{db}` is not initialized")
        self.provider: BaseProvider = provider

    async def execute(self, query: str, args: ty.Sequence[ty.Any] = ()):
        """
        Executes SQL query.
        :param query: SQL statement.
        :param args: statement params.
        :returns: cursor instance.
        """
        return await self.provider.execute(query, args)

    async def create_tables(self) -> None:
        """
        Creates all tables.
        """
        for model_name, signature in self.signatures.items():
            query = self.provider.prepare_create_table_query(
                model_name,
                {
                    field.name: (
                        field.python_type,
                        field.unique,
                        field.sql_type,
                    )
                    for field in signature.fields
                },
            )
            await self.provider.execute(query)

    async def insert(
        self, objs: Models | list[Models], /
    ) -> Models | list[Models]:
        """
        Inserts obj to bd.
        :param objs: objs to insert.
        :returns: The same object, but with an identifier in the database.
        """
        if not (return_list := isinstance(objs, list)):
            objs = [objs]
        if len(set(type(x) for x in objs)) != 1:
            raise ValueError("objects must be same types")
        signature = self.signatures[objs[0].__class__.__name__]
        field_names = [
            field.name for field in signature.fields if field.name != "id"
        ]
        values = []
        for obj in objs:
            for field in signature.fields:
                if field.name != "id":
                    values.append(getattr(obj, field.name))
        query = self.provider.prepare_insert_query(
            signature.name, field_names, len(objs)
        )
        obj_ids = await self.provider.execute_insert_query(query, values)
        for obj, obj_id in zip(objs, obj_ids):
            obj.id = obj_id
        return objs if return_list else objs[0]

    def _prepare_select_query(
        self,
        model_name: str,
        join: Join[Models] | None = None,
        where: Operator | None = None,
        order_by: (
            Field | InvertedField | tuple[Field | InvertedField, ...] | None
        ) = None,
        limit: int | None = None,
        offset: int = 0,
    ) -> str:
        if order_by and not isinstance(order_by, tuple):
            order_by = (order_by,)
        return self.provider.prepare_select_query(
            model_name,
            join=str(join) if join else None,
            where=str(where) if where is not None else None,
            order_by=(
                tuple(
                    (str(x) if isinstance(x, Field) else (str(x), True))
                    for x in order_by
                )
                if order_by is not None
                else None
            ),
            limit=limit,
            offset=offset,
        )

    def _convert_data(
        self,
        model: ty.Type[Models],
        data: tuple[ty.Any, ...],
        join: Join[Models] | None = None,
    ):
        """
        Converts raw data from db to model.
        """
        signature = self.signatures[model.__name__]
        if join:
            sep = len(signature.fields)
            return (
                self._convert_data(model, data[:sep]),
                self._convert_data(join.model, data[sep:]),
            )
        elif set(data) == {None}:
            return None

        if len(data) != len(signature.fields):
            raise ValueError(
                f"Model {signature.name} have {len(signature.fields)} fields, "
                f"but {len(data)} given"
            )

        kwargs: dict[str, ty.Any] = {}
        for field, value in zip(signature.fields, data):
            kwargs[field.name] = self.provider.convert_value(
                value, field.python_type
            )

        return model(**kwargs)

    async def fetchone(
        self,
        model: ty.Type[Models],
        *,
        join: Join[Models] | None = None,
        where: Operator | None = None,
        order_by: (
            Field | InvertedField | tuple[Field | InvertedField, ...] | None
        ) = None,
        limit: int | None = None,
        offset: int = 0,
    ):
        """
        Fetches one row from db.
        :param model: model to fetch.
        :param join: join statement.
        :param where: filtering statement.
        :param order_by: field for sorting.
        :param limit: count of rows to fetch.
        :param offset: offset of rows to fetch.
        :returns: one model or None.
        """
        query = self._prepare_select_query(
            model.__name__, join, where, order_by, limit, offset
        )
        if data := await self.provider.fetchone(
            query, where.get_values() if where is not None else ()
        ):
            return self._convert_data(model, data, join)

    async def fetchall(
        self,
        model: ty.Type[Models],
        *,
        join: Join[Models] | None = None,
        where: Operator | None = None,
        order_by: (
            Field | InvertedField | tuple[Field | InvertedField, ...] | None
        ) = None,
        limit: int | None = None,
        offset: int = 0,
    ):
        """
        Fetches all rows from db.
        :param model: model to fetch.
        :param join: join statement.
        :param where: filtering statement.
        :param order_by: field for sorting..
        :param limit: count of rows to fetch.
        :param offset: offset of rows to fetch.
        :returns: list of model or empty list.
        """
        query = self._prepare_select_query(
            model.__name__, join, where, order_by, limit, offset
        )
        data = await self.provider.fetchall(
            query, where.get_values() if where is not None else ()
        )
        return (
            [self._convert_data(model, obj, join) for obj in data]
            if data
            else []
        )

    async def save(self, obj: Models, /) -> None:
        """
        Saves obj to db.
        :param obj: obj to save.
        """
        if not (changed_field_names := get_changed_attributes(obj)):
            return
        model = obj.__class__
        return await self.update(
            model,
            {
                getattr(model, field_name): getattr(obj, field_name)
                for field_name in changed_field_names
                if field_name != "id"
            },
            where=model.id == obj.id,
        )

    async def update[T](
        self,
        model: ty.Type[Models],
        fields: dict[Field[T], T | MathOperator[T]],
        *,
        where: Operator | None = None,
    ) -> None:
        """
        Executes update query.
        :param model: model to update.
        :param fields: fields to set.
        :param where: filtering statement.
        """
        query = self.provider.prepare_update_query(
            model.__name__,
            {
                field.name: value.sign
                if isinstance(value, MathOperator)
                else "="
                for field, value in fields.items()
            },
            where=str(where) if where is not None else None,
        )
        await self.execute(
            query,
            (
                *(
                    value.value if isinstance(value, MathOperator) else value
                    for value in fields.values()
                ),
                *(where.get_values() if where is not None else ()),
            ),
        )

    async def delete(
        self, model: ty.Type[Models], *, where: Operator | None
    ) -> None:
        """
        Deletes row from db.
        :param model: model from which to delete.
        :param where: filtering statement.
        """
        query = self.provider.prepare_delete_query(
            model.__name__, where=str(where) if where is not None else None
        )
        await self.execute(
            query, where.get_values() if where is not None else ()
        )

    async def drop_table(self, model: ty.Type[Models], /) -> None:
        """
        Drops table from db.
        :param model: model to drop.
        """
        query = self.provider.prepare_drop_table_query(model.__name__)
        await self.execute(query)



import importlib
import re
import typing as ty

from .base import BaseProvider


def get_provider(db_url: str) -> ty.Type[BaseProvider]:
    """
    Imports suitable provider class and returns it.
    :param db_url: path to db.
    :returns: provider class.
    """
    if not (
        match := re.fullmatch(
            r"(?P<provider>.+?)(\+(?P<library>.+?))?://(?P<db>.+)", db_url
        )
    ):
        raise ValueError(
            "Invalid db url. pass url like `sqlite+aiosqlite://db.path`"
        )
    provider = match.group("provider")
    library = match.group("library")
    if provider == "postgres":
        provider = "postgresql"

    try:
        module = importlib.import_module(
            f".{library}_provider", package=f"{__name__}.{provider}_providers"
        )
    except ModuleNotFoundError as err:
        raise ValueError(f"`{provider}` or `{library}` not supported") from err

    return getattr(module, f"{library.capitalize()}Provider")
